package wam;

use strict;
use Carp;
use vars qw($VERSION @ISA @EXPORT_OK $AUTOLOAD);

require Exporter;
require DynaLoader;
require AutoLoader;

@ISA = qw(Exporter DynaLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT_OK = qw(
	WAM_ROW
	WAM_COL
);
$VERSION = '0.01';

sub AUTOLOAD {
    # This AUTOLOAD is used to 'autoload' constants from the constant()
    # XS function.  If a constant is not found then control is passed
    # to the AUTOLOAD in AutoLoader.

    my $constname;
    ($constname = $AUTOLOAD) =~ s/.*:://;
    croak "& not defined" if $constname eq 'constant';
    my $val = constant($constname, @_ ? $_[0] : 0);
    if ($! != 0) {
	if ($! =~ /Invalid/) {
	    $AutoLoader::AUTOLOAD = $AUTOLOAD;
	    goto &AutoLoader::AUTOLOAD;
	}
	else {
		croak "Your vendor has not defined wam macro $constname";
	}
    }
    no strict 'refs';
#    *$AUTOLOAD = sub () { $val };
    *$AUTOLOAD = sub { $val };
    goto &$AUTOLOAD;
}

bootstrap wam $VERSION;

# Preloaded methods go here.

# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

B<wam> - WAM library module for perl

=head1 SYNOPSIS

B<use wam>;

=head1 DESCRIPTION

このモジュール (以下，WAM モジュール) は GETAssoc が提
供する C 言語ライブラリを Perl 言語から使うためのものである。 

=head2 関数一覧

WAM モジュールでは以下の関数が提供されている。

=over 4

=item B<wam::init>(F<$projectroot>)

WAM モジュールを初期化する。
F<$projectroot> には GETA 環境がインストールされているディレクトリの絶
対パス名を指定する。
F<$projectroot> に空文字列を指定した場合は，環境変数 GETAROOT の値が使
用される。
初期化に成功した場合は 0，失敗した場合は -1 を返す。

=item B<wam::open>(F<$handle>)

ハンドルが F<$handle> である WAM をオープンし，その識別子 (正の整数) 
を返す。
オープン後，ユーザはこの識別子を用いて WAM にアクセスすることができる。
オープンに失敗した場合は 0 を返す。

=item B<wam::close>(F<$d>)

識別子 F<$d> に関連付けられた WAM をクローズする。


=item B<wam::id2name>(F<$d>, F<$mode>, F<$id>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側の名前表を引き，与えられた ID F<$id> に対応する name を返す。
対応する name がない場合は空文字列を返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::name2id>(F<$d>, F<$mode>, F<$name>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側の名前表を引き，与えられた name F<$name> に対応する ID を返す。 
対応する ID がない場合は 0 を返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::idl2namel>(F<$d>, F<$mode>, F<$idl>)

B<wam::id2name> の配列対応版。
ID の配列へのリファレンス F<$idl> をうけとり，それぞれの ID に対応する 
name を要素としてもつ配列へのリファレンスを返す。
対応する name がない ID に対しては空文字列が割り当てられる。

F<$d> と F<$mode> は B<wam::id2name> と同じ意味である。

=item B<wam::namel2idl>(F<$d>, F<$mode>, F<$namel>)

B<wam::name2id> の配列対応版。
name の配列へのリファレンス F<$namel> をうけとり，それぞれの name に対
応する ID を要素としてもつ配列へのリファレンスを返す。 
対応する ID がない name に対しては 0 が割り当てられる。

F<$d> と F<$mode> は B<wam::name2id> と同じ意味である。

=item B<wam::size>(F<$d>, F<$mode>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側の要素数を返す。

F<$mode> に &wam::WAM_ROW を指定すると WAM の行数を，&wam::WAM_COL を
指定すると WAM の列数を知ることができる。


=item B<wam::freq_sum>(F<$d>, F<$mode>, F<$id>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側 の F<$id> ベクタの要素の値の総和を返す。 
存在しない ID を与えた場合は -1 を返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::total_freq_sum>(F<$d>, F<$mode>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側 (行または列) のすべての行 (列) ベクタの要素の値の総和を返す。
すなわち，識別子 F<$d> に関連付けられた WAM (厳密には XR) に含まれる要
素の値の総和を返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::elem_num>(F<$d>, F<$mode>, F<$id>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側 (行または列) の F<$id> 行 (列) ベクタ中の 0 でない要素数を
返す。
存在しない ID を与えた場合は -1 を返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::max_elem_num>(F<$d>, F<$mode>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側 (行または列) の 行 (列) ベクタ中の 0 でない要素数の最大値を
返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::total_elem_sum>(F<$d>, F<$mode>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側 (行または列) のすべての行 (列) ベクタ中の 0 でない要素数の
総和を返す。
すなわち，識別子 F<$D> に関連付けられた WAM (厳密には XR) に含まれる 0 
でない要素数を返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::get_vec>(F<$d>, F<$mode>, F<$id>)

識別子 F<$d> に関連付けられた WAM のうち，F<$mode> で指
定された側 (行または列) の F<$id> 行 (列) ベクタ を返す。

戻り値は，id, freq, name の 3 つのキーをもつ連想配列へのリファレンスの
配列へのリファレンスである。
存在しない ID を与えた場合は空文字列を返す。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::text2vec>(F<$text>, F<$d>, F<$mode>, F<$stemmer>)

与えられた文字列 F<$text> を，名前が F<$stemmer> である形態素解析器で単
語に分割する。
その結果を name であるとして，識別子 F<$d> に関連付けられた WAM 
のうち，F<$mode> で指定された側の対応表 (行 CW または列 CW) 
を引き，その name に対応する ID を割り当てる。
対応する ID がない単語は無視される。

戻り値は，id, freq, name の 3 つのキーをもつ連想配列へのリファレンスの
配列へのリファレンス (wam::get_vec と同じ形式)である。
対応する ID をもつ単語がひとつもない場合は，長さ 0 の配列へのリファレ
ンスが返される。

freq には，その ID の検索要求 F<$text> 中での出現回数 (TF) がセットされる。

F<$mode> には， &wam::WAM_ROW (行) または &wam::WAM_COL (列) を指定す
る。

=item B<wam::wsh>(F<$r>, F<$d1>, F<$mode>, F<$weight_type>, F<$max>, F<$limit>, F<$d2>) 

連想検索を行なう。
検索要求(詳細は後述) F<$r> 中の ID (行 ID または列 ID) に対応するベクタに現れる ID
(列 ID または行 ID) を最大 F<$max> 個集め，その結果を id, TF, TF_d,
DF, DF_d, weight, name の 7 つのキーをもつ連想配列へのリファレンスの配
列へのリファレンスとして返す。
id，TF，TF_d，DF，DF_d，weight は syminfo 型の同名のメンバに対応している。
詳しくは B<wsh>(3) を参照のこと。

検索要求F<$r>は, ハッシュのリストであり, 各要素がひとつのベクタに対応する.
各要素は 少なくとも ID 持たなければならなず, また TF_d, weight, TF を持つ
ことができる.
ID は以下のいずれかの方法で与えなければならない.
	1. キー: v  値: ベクタ(get_vec で取得したもの)
	2. キー: name 値: 記事名もしくは単語綴り
	3. キー: id 値: 記事名もしくは単語綴りを id2name で ID に変換したもの
複数指定したときには前のものが優先される.
TF_d は ID が検索要求中に現れた回数を指定する. キーは TF_d である. 省略すると 1 になる.
weight は ID の検索要求中での重要度を指定する. キーは weight である. 省略すると 1 になる.
TF_d, weight は類似度によっては無視される.
キー TF は予約されている.
上記以外のキーは無視される. (エラーにはならない)

結果は，類似度 F<$weight_type> を用いて計算された weight の値の高い順
に配列に格納される。
使用可能な類似度のリストは B<wam::get_weight_types> で得ることができる。

検索要求 F<$r> が行 ID の場合は F<$mode> に &wam::WAM_ROW (行) を指定
し，列 ID の場合は &wam::WAM_COL (列) を指定する。

識別子 F<$d1> には検索要求 F<$r> が含まれる WAM を示す識別子を指定する。
識別子 F<$d2> には検索対象となる WAM を示す識別子を指定する。
同一文書内で連想検索を行なう場合は F<$d1> == F<$d2>，異文書間で行なう
場合は F<$d1> != F<$d2> である。

F<$limit> に正の整数を指定すると，検索要求 F<$r> 中の ID のうち，識別
子 F<$d2> に関連付けられた WAM における TF 値が  F<$limit> 以上の ID 
を連想計算に使用しない。
0 を指定すると検索要求中のすべての ID を連想計算に使用する。
この値を適切に設定することで検索結果が膨大な数になるのを防ぐことができ
る。

検索要求 F<$r> が単語リストの場合は結果として文書リストが得られ，文書
リストの場合は単語リストが得られる。

検索要求 F<$r> は name, id, TF, TD_d, weight の 5 つのキーをもつ
連想配列へのリファレンスの配列へのリファレンスである。

name キーは特に指定する必要はなく，指定しても無視される。
id キーを省略した場合は，0 を指定したものとして解釈される。
TF キー，TF_d キー，weight キーを省略した場合は，それぞれ，0, 0, 0.0 
を指定したものとして解釈され，こららの値を使用する類似度を使用している
場合は所望の結果を得ることができない。

キー attr は廃止された.

=item B<wam::wsa>(F<$r>, F<$d1>, F<$mode>, F<$max>, F<$limit>, F<$d2>)

B<wam::wsa> は廃止された

=item B<wam::wshb>(F<$r>, F<$b>, F<$d1>, F<$mode>, F<$weight_type>, F<$max>, F<$limit>, F<$d2>) 

B<wam::wshb> は wsh の ブーリアンフィルタ付き版である.
F<$b> にブール式を渡すことができる.

=item B<wam::wshf>(F<$r>, F<$f>, F<$d1>, F<$mode>, F<$weight_type>, F<$max>, F<$limit>, F<$d2>) 

B<wam::wshb> は wsh の 全文一致検索フィルタ付き版である.
F<$f> に全文一致検索式を渡すことができる.

=item B<wam::getprop>(F<$d>, F<$dir>, F<$basename>, F<$id>)

識別子 F<$d> に関連付けられた WAM  に属する F<$dir> の方向の補助ファイ
ル F<$basename> (タイトルファイルなど) をアクセスし，与えられた 
ID F<$id> に対応する値を返す。失敗した場合は空文字列を返す。

=item B<wam::getfss>(F<$d>, F<$id>, F<$segid>)

識別子 F<$d> に関連付けられた WAM  に属する全文一致ファイルをアクセスし，
与えられた
ID F<$id> の セグメント F<$segid> を返す。失敗した場合は空文字列を返す。

=item B<wam::get_weight_types>()

libae で定義されている類似度の名前の一覧を返す。
$GETAROOT/include/geta/weight_types.h に定義されている文字列の配列 
weight_types の要素を perl の配列に順にセットし，その配列へのリファレ
ンスを戻り値として返す。

=item B<wam::ci_value>(F<$handle>, F<$attr>)

B<wam::ci_value> は廃止された

=item B<wam::get_last_nd>()

B<wam::wsh> または B<wam::wsa> で連想検索された ID の総数を返す。

=item B<wam::set_prob_mode>(F<$handle>, F<$e>)

ハンドルが F<$handle> である WAM の確率的動作モードを有効にする.
許容エラーレートは 1/F<$e> にセットされる.
F<$e> に 0 を指定すると確率的動作モードが解除される.
ハンドルの代わりにショートネームを使ってもよい。

=back

=head1 TECHNICAL NOTE

以下，技術的な詳細について補足説明しておく。

=head2 データ型について

Perl では C 言語のような構造体を扱うことができない。
そのため，WAM モジュールでは連想配列とリファレンスを組み合わせて構造体
に相当するデータ構造を表現している。

libwam で用いられている xr_vec 型に対応するデータ構造は WAM モジュール 
にはない。

libwam で用いられている xr_elem 型は id, freq, name の 3 つのキーをも
つ連想配列として表現される。
id と freq は xr_elem 構造体の同名メンバに対応する。
name には，その連想配列の id に対応する name が 割り当てられる。

libae で用いられている syminfo 型に対応するデータ構造は id, TF, TF_d,
DF, DF_d, weight, name の 7 つのキーをもつ連想配列である。
id，TF，TF_d，DF，DF_d，weight は syminfo 型の同名のメンバに対応する。
name には，その連想配列の id に対応する name が 割り当てられる。

=head2 関数の対応関係

WAMモジュールが提供する関数と C 言語ライブラリが提供する関数の対応関係
は以下のとおりである。

 WAMモジュール		C 言語ライブラリ
  wam::init		 wam_init
  wam::open		 wam_open
  wam::close		 wam_close
  wam::id2name		 wam_id2name
  wam::name2id		 wam_name2id
  wam::idl2namel	 ---
  wam::namel2idl	 ---
  wam::size		 wam_size
  wam::freq_sum		 wam_freq_sum
  wam::total_freq_sum	 wam_total_freq_sum
  wam::elem_num		 wam_elem_num
  wam::max_elem_num	 wam_max_elem_num
  wam::total_elem_num	 wam_total_elem_num
  wam::get_vec		 wam_get_vec
  wam::text2vec		 text2vec
  wam::wsh		 wsh => assv
  wam::wshb		 bex_wsh => assv
  wam::wshf		 fss_wsh => assv
  wam::getprop		 wam_prop_gets
  wam::getfss		 wam_fss_gets
  wam::get_weight_types	 ---
  wam::get_last_nd	 ---
  wam::set_opt		 wam_set_opt

  以下，対応する関数間で仕様が異なるものに関して補足説明をする。

=over 4

=item B<wam::open> と B<wam_open>

B<wam_open> は WAM * 型のポインタを返すのに対し，B<wam::open>は小さい
正の整数を返す。

=item B<wam::get_vec> と B<wam_get_vec>

B<wam_get_vec> では xr_vec 型の構造体を用いて行 (列) ベクタを表現して
いるのに対し，B<wam::get_vec> では ID で示される行 (列) ベクタのベ
クタ部分だけ (xr_vec 型の構造体のメンバである xr_elem 型の要素に相当) 
を id, freq, name の 3 つのキーをもつ連想配列として表現している。
詳しくは B<xr_get_vec>(3) を参照のこと。

=item B<wam::text2vec> と B<text2vec>

B<text2vec> は xr_vec 型へのポインタを返すのに対し，B<wam::text2vec> は 
name, id, TF, TD_d, weight の 5 つのキーをもつ連想配列へのリファレンス
の配列へのリファレンスを返す。

=item B<wam::wsh> と B<wsh> または B<assv>

B<wsh> では，検索要求と検索結果のいずれも syminfo 型へのポインタである
のに対し，B<wam::wsh> では，検索要求が name, id, TF, TD_d, weight,
の 5 つのキーをもつ連想配列へのリファレンスの配列へのリファレンス，
検索結果が id, TF, TF_d, DF, DF_d, weight, name の 7 つのキーをもつ連
想配列へのリファレンスの配列へのリファレンスとなっている。

B<wsh>では検索要求の長さを渡す必要があるが，B<wam::wsh> ではその必要が
ない。

B<wsh>は検索結果の配列の要素数も返すが，B<wam::wsh> は返さない。

B<wsh>では，cookie を指定することができるが，B<wam::wsh> ではできない 
(内部的には NULL で固定)。

=item B<wam::getprop> と B<wam_prop_gets>

B<wam_prop_gets> はデータの格納領域へのポインタを返すのに対し，B<wam::getprop> 
は，そのポインタが指す文字列を SV に変換したものを返す。

=item B<wam::set_opt>

B<wam_set_opt> は第3引数にデータの格納領域へのポインタを渡すのに対し，B<wam::set_opt> 
は，そのポインタが指すべきデータを渡す.
現在の実装では, B<wam_set_opt> の第3引数は int のみなので,
この仕様で問題無いのだが, 将来, そうでなくなった場合に,
B<wam::set_opt> の仕様を変更する必要があるだろう.

=back

=head1 PERL VERSION

WAM モジュールを利用するためには Perl version 5.8.8 以降が必要であ
る。

=head1 SEE ALSO

B<wam>(3), B<xr>(3), B<cw>(3)

=head1 AUTHOR

 今一 修 <imaichi@harl.hitachi.co.jp>
 西岡 真吾 <nis@harl.hitachi.co.jp> 

=cut
