/* static char rcsid[] = "$Id: wam.xs,v 1.10 2007/03/29 07:37:38 nis Exp $"; */

#include <config.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/uio.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <math.h>
#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include "nwam.h"

#define nelems(e)	(sizeof (e) / sizeof *(e))

#define	sizeof_dxr_vec(n)	(offsetof(struct xr_vec, elems[(n)]))

typedef struct syminfo syminfo;

static SV *vec2rh(struct xr_vec const *, WAM *, int);
static struct xr_vec *rh2vec(AV *, WAM *, int);
static SV *s2rah(syminfo *, size_t, WAM *, int);
static syminfo *rah2s(AV *, WAM *, int, ssize_t *);
static syminfo *rah2s(AV *, WAM *, int, ssize_t *);
static struct bxue_t *rah2bxue_t(AV *, WAM *, int, ssize_t *);
static struct fss_query *raah2fss_query(AV *);
static int rah2fss_con_query(AV *, struct fss_con_query *);

static int xs_wam_open(char *);
static int xs_wam_close(int);
static WAM *xs_wam_desc(int);

static size_t iwl(syminfo *, size_t, size_t, WAM *, int);
static SV *wsh0(SV *, SV *, SV *, int, int, int, size_t, size_t, int);
static SV *ncsb0(SV *, int, int, int, int, size_t, int32_t, size_t);
static SV *idl2namel0(int, int, SV *);
static SV *namel2idl0(int, int, SV *);
static void free_vecs(syminfo *, size_t);
// fssx, fss_wsh, csb

static ssize_t lastnd = 0;

static char const *weight_types[] = WEIGHT_TYPE_NAMES;
static char const *cs_types[] = CS_TYPE_NAMES;

#define	sym_compar	n_sym_compar

#define	my_hv_store(nh, key, sv) hv_store((nh), #key, sizeof #key-1, (sv), 0)
#define	my_hv_fetch(h, key) hv_fetch((h), #key, sizeof #key-1, 0)

static int
not_here(char *s)
{
	croak("%s not implemented on this architecture", s);
	return -1;
}

static int
constant(char *name, int arg)
{
	size_t i;
	errno = 0;

	for (i = 0; i < nelems(weight_types); i++) {
		if (strEQ(name, weight_types[i])) {
			return i;
		}
	}

	for (i = 0; i < nelems(cs_types); i++) {
		if (strEQ(name, cs_types[i])) {
			return i;
		}
	}

	switch (*name) {
	case 'A':
		break;
	case 'B':
		break;
	case 'C':
		break;
	case 'D':
		break;
	case 'E':
		break;
	case 'F':
		break;
	case 'G':
		break;
	case 'H':
		break;
	case 'I':
		break;
	case 'J':
		break;
	case 'K':
		break;
	case 'L':
		break;
	case 'M':
		break;
	case 'N':
		break;
	case 'O':
		break;
	case 'P':
		break;
	case 'Q':
		break;
	case 'R':
		break;
	case 'S':
		break;
	case 'T':
		break;
	case 'U':
		break;
	case 'V':
		break;
	case 'W':
	if (strEQ(name, "WAM_ROW"))
#if defined WAM_ROW
		return WAM_ROW;
#else
		goto not_there;
#endif
	if (strEQ(name, "WAM_COL"))
#if defined WAM_COL
		return WAM_COL;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_OR"))
#if defined WSH_OR
		return WSH_OR;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_AND"))
#if defined WSH_AND
		return WSH_AND;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NOT"))
#if defined WSH_NOT
		return WSH_NOT;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_ORG"))
#if defined WSH_ORG
		return WSH_ORG;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_CE"))
#if defined WSH_NF_CE
		return WSH_NF_CE;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_NOT"))
#if defined WSH_NF_NOT
		return WSH_NF_NOT;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_OR"))
#if defined WSH_NF_OR
		return WSH_NF_OR;
#else
		goto not_there;
#endif
	if (strEQ(name, "WSH_NF_DE"))
#if defined WSH_NF_DE
		return WSH_NF_DE;
#else
		goto not_there;
#endif
		break;
	case 'X':
		break;
	case 'Y':
		break;
	case 'Z':
		break;
	}
	errno = EINVAL;
	return 0;

	not_there:
	errno = ENOENT;
	return 0;
}

static SV *
vec2rh(vec, w, mode)
WAM *w;
struct xr_vec const *vec;
{
	AV *na;
	struct xr_elem const *ep, *ee;
	size_t e;

	e = vec->elem_num;
	na = newAV();
	av_extend(na, e);
	for (ep = vec->elems, ee = ep + e; ep < ee; ep++) {
		HV *nh;
		SV *sv;
		char const *p;
		nh = newHV();

		sv = newSViv(ep->id);
		my_hv_store(nh, id, sv);

		p = wam_id2name(w, mode, ep->id);
		sv = newSVpv(p, 0);
		my_hv_store(nh, name, sv);

		sv = newSViv(ep->freq);
		my_hv_store(nh, freq, sv);

		sv = newRV_noinc((SV *)nh);
		av_push(na, sv);
	}
	return newRV_noinc((SV *)na);
}

static struct xr_vec *
rh2vec(a, w, mode)
AV *a;
WAM *w;
{
	size_t len;
	size_t i;
	struct xr_vec *v;
	struct xr_elem *ep;

	len = av_len(a) + 1;
	if (!(v = malloc(sizeof_dxr_vec(len)))) {
		perror("calloc");
		return NULL;
	}
	v->freq_sum = 0;
	for (i = 0, ep = v->elems; i < len; i++, ep++) {
		SV **sp;

		sp = av_fetch(a, i, 0);
		if (sp && SvROK(*sp) && SvTYPE(SvRV(*sp)) == SVt_PVHV) {
			HV *h;
			SV **s;
			h = (HV *)SvRV(*sp);

			s = my_hv_fetch(h, name);
			if (s && SvPOK(*s)) {
				char *name;
				STRLEN l;
				name = SvPV(*s, l);
				ep->id = wam_name2id(w, mode, name);
			}
			else {
				s = my_hv_fetch(h, id);
				ep->id = s ? SvIV(*s) : 0;
			}

			s = my_hv_fetch(h, freq);
			ep->freq = s ? SvIV(*s) : 0;
			v->freq_sum += ep->freq;
		}
		else {
			ep->id = 0;
			ep->freq = 0;
		}
	}
	v->elem_num = len;
	return v;
}

static SV *
s2rah(dp, nd, w, dir)
syminfo *dp;
size_t nd;
WAM *w;
{
	AV *na;
	ssize_t i;

	na = newAV();
	av_extend(na, nd);
	for (i = 0; i < nd; i++) {
		HV *nh;
		SV *sv;
		char const *p;
		nh = newHV();

		sv = newSViv(dp[i].id);
		my_hv_store(nh, id, sv);

		p = wam_id2name(w, dir, dp[i].id);
		sv = newSVpv(p, 0);
		my_hv_store(nh, name, sv);

		sv = newSVnv(dp[i].weight);
		my_hv_store(nh, weight, sv);

		sv = newSViv(dp[i].TF_d);
		my_hv_store(nh, TF_d, sv);

		sv = newSViv(dp[i].TF);
		my_hv_store(nh, TF, sv);

		sv = newSViv(dp[i].DF_d);
		my_hv_store(nh, DF_d, sv);

		sv = newSViv(dp[i].DF);
		my_hv_store(nh, DF, sv);

		/* assv never returns dp[*].v. just ignore member `v'. */

		sv = newRV_noinc((SV *)nh);
		av_push(na, sv);
	}
	return newRV_noinc((SV *)na);
}

static syminfo *
rah2s(a, w, mode, lenp)
AV *a;
WAM *w;
ssize_t *lenp;
{
	size_t len;
	size_t i;
	syminfo *q, *qp;

	len = av_len(a) + 1;
	if (!(q = calloc(len, sizeof *q))) {
		perror("calloc");
		return NULL;
	}
	for (i = 0, qp = q; i < len; i++, qp++) {
		SV **sp;

		sp = av_fetch(a, i, 0);
		if (sp && SvROK(*sp) && SvTYPE(SvRV(*sp)) == SVt_PVHV) {
			HV *h;
			SV **s;
			h = (HV *)SvRV(*sp);

			qp->id = 0;
			qp->v = NULL;
			s = my_hv_fetch(h, v);
			if (s && SvROK(*s) && SvTYPE(SvRV(*s)) == SVt_PVAV) {
				qp->v = rh2vec((AV *)SvRV(*s), w, WAM_REVERT_DIRECTION(mode));
			}
			else {
				s = my_hv_fetch(h, name);
				if (s && SvPOK(*s)) {
					char *name;
					STRLEN l;
					name = SvPV(*s, l);
					qp->id = wam_name2id(w, mode, name);
				}
				else {
					s = my_hv_fetch(h, id);
					qp->id = s ? SvIV(*s) : 0;
				}
			}

			if (s = my_hv_fetch(h, TF_d)) {
				qp->TF_d = SvIV(*s);
			}
			else if (s = my_hv_fetch(h, freq)) {
				qp->TF_d = SvIV(*s);
			}
			else {
				qp->TF_d = 1;
			}

			s = my_hv_fetch(h, TF);
			qp->TF = s ? SvIV(*s) : 0;

			s = my_hv_fetch(h, weight);
			qp->weight = s ? SvNV(*s) : 0.0;
		}
		else {
			qp->id = 0;
			qp->TF_d = 0;
			qp->TF = 0;
			qp->weight = 0;
			qp->v = NULL;
		}
	}
	*lenp = len;
	return q;
}

static struct bxue_t *
rah2bxue_t(a, w, mode, lenp)
AV *a;
WAM *w;
ssize_t *lenp;
{
	size_t len;
	size_t i;
	struct bxue_t *q, *qp;

	len = av_len(a) + 1;
	if (!(q = calloc(len, sizeof *q))) {
		perror("calloc");
		return NULL;
	}
	for (i = 0, qp = q; i < len; i++, qp++) {
		SV **sp;

		sp = av_fetch(a, i, 0);
		if (sp && SvROK(*sp) && SvTYPE(SvRV(*sp)) == SVt_PVHV) {
			HV *h;
			SV **s;
			h = (HV *)SvRV(*sp);
			s = my_hv_fetch(h, name);
			if (s && SvPOK(*s)) {
				char *name;
				STRLEN l;
				name = SvPV(*s, l);
				qp->id = wam_name2id(w, mode, name);
				qp->type = '"';
			}
			else {
				s = my_hv_fetch(h, id);
				if (s && SvIOK(*s)) {
					qp->id = SvIV(*s);
					qp->type = '"';
				}
				else {
					qp->id = 0;
					s = my_hv_fetch(h, type);
					qp->type = s ? SvIV(*s) : 0;
				}
			}
		}
		else {
			qp->type = 0;
			qp->id = 0;
		}
	}
	*lenp = len;
	return q;
}

static struct fss_query *
raah2fss_query(a)
AV *a;
{
	size_t i;
	size_t len;
	struct fss_con_query *cq;
	struct fss_query *r;

	len = av_len(a) + 1;

	if (!(r = calloc(1, sizeof *r))) {
		perror("calloc");
		return NULL;
	}
	if (!(cq = calloc(len, sizeof *cq))) {
		perror("calloc");
		return NULL;
	}
	for (i = 0; i < len; i++) {
		SV **sp;

		sp = av_fetch(a, i, 0);
		if (sp && SvROK(*sp) && SvTYPE(SvRV(*sp)) == SVt_PVAV) {
			AV *av;
			av = (AV *)SvRV(*sp);
			if (rah2fss_con_query(av, &cq[i]) == -1) {
				return NULL;
			}
		}
	}
	r->query = cq;
	r->nq = len;
	return r;
}

static int
rah2fss_con_query(a, cq)
AV *a;
struct fss_con_query *cq;
{
	size_t len;
	size_t i;
	struct fss_simple_query *q, *qp;

	len = av_len(a) + 1;
	if (!(q = calloc(len, sizeof *q))) {
		perror("calloc");
		return -1;
	}
	for (i = 0, qp = q; i < len; i++, qp++) {
		SV **sp;

		qp->negativep = 0;
		qp->pattern = NULL;
		qp->options = 0;
		qp->segments = ANYSEG();
		qp->arts = NULL;
		qp->narts = qp->total = 0;

		sp = av_fetch(a, i, 0);
		if (sp && SvROK(*sp) && SvTYPE(SvRV(*sp)) == SVt_PVHV) {
			HV *h;
			SV **s;
			h = (HV *)SvRV(*sp);
			s = my_hv_fetch(h, pattern);
			if (s && SvPOK(*s)) {
				char *name;
				STRLEN l;
				name = SvPV(*s, l);
				qp->pattern = strdup(name);
			}
			s = my_hv_fetch(h, negativep);
			if (s && SvIOK(*s)) {
				qp->negativep = SvIV(*s);
			}
			s = my_hv_fetch(h, segments);
			if (s && SvPOK(*s)) {
				char *sspec;
				STRLEN l;
				sspec = SvPV(*s, l);
				if (parse_segs(sspec, &qp->segments) == -1) {
					return -1;
				}
			}
			s = my_hv_fetch(h, options);
			if (s && SvPOK(*s)) {
				char *opts;
				STRLEN l;
				opts = SvPV(*s, l);
				if (parse_opts(opts, &qp->options) == -1) {
					return -1;
				}
			}
		}
	}
	cq->q = q;
	cq->n = len;
	return 0;
}

#undef	my_hv_store
#undef	my_hv_fetch

static WAM *dt[256];

static int
xs_wam_open(handle)
char *handle;
{
	int d;
	static int initialized = 0;
	if (!initialized) {
		for (d = 0; d < nelems(dt); d++) {
			dt[d] = NULL;
		}
		initialized = 1;
	}
	for (d = 1; d < nelems(dt); d++) {
		if (!dt[d]) {
			if ((dt[d] = wam_open(handle)) != NULL) {
				return d;
			}
			else {
				return 0;
			}
		}
	}
	return 0;
}

static int
xs_wam_close(d)
{
	if (d <= 0 || nelems(dt) <= d || dt[d] == NULL) {
		return -1;
	}
	wam_close(dt[d]);
	dt[d] = NULL;
	return 0;
}

static WAM *
xs_wam_desc(d)
{
	if (d <= 0 || nelems(dt) <= d) {
		return NULL;
	}
	return dt[d];
}

static size_t
iwl(terms, nkwords, tak_iw_limit, w, dir)
syminfo *terms;
size_t nkwords, tak_iw_limit;
WAM *w;
{
	size_t i, j;

	if (nkwords <= 0) {
		return nkwords;
	}
	/* note: terms[j..nkwords] ni gomi ga nokoru. */
	for (i = 0, j = 0; i < nkwords; i++) {
		if (wam_elem_num(w, dir, terms[i].id) <= tak_iw_limit) {
			terms[j++] = terms[i];
		}
	}
	return MAX(j, 1);
}

static SV *
wsh0(s, e, f, d, mode, weight_type, maxnd, tak_iw_limit, target_wam)
SV *s;
SV *e;
SV *f;
int d;
int mode;
int weight_type;
size_t maxnd;
size_t tak_iw_limit;
int target_wam;
{
	AV *a;
	syminfo *q = NULL;
	struct bxue_t *qe = NULL;
	struct fss_query *fq = NULL;
	ssize_t len, elen;
	syminfo *dp = NULL;
	int32_t nd, total;
	WAM *w, *wt;
	SV *r;

	if (!(SvROK(s) && SvTYPE(SvRV(s)) == SVt_PVAV)) {
		return NULL;
	}
	if (e && !(SvROK(e) && SvTYPE(SvRV(e)) == SVt_PVAV)) {
		return NULL;
	}
	if (f && !(SvROK(f) && SvTYPE(SvRV(f)) == SVt_PVAV)) {
		return NULL;
	}
	a = (AV *)SvRV(s);
	nd = maxnd;
	if (!(w = xs_wam_desc(d))) {
		return NULL;
	}
	if (!(q = rah2s(a, w, mode, &len))) {
		return NULL;
	}

	if (e) {
		a = (AV *)SvRV(e);
		if (!(qe = rah2bxue_t(a, w, mode, &elen))) {
			free_vecs(q, len);
			free(q);
			return NULL;
		}
		dp = bex_wsh(q, len, w, mode, weight_type, &nd, &total, qe, elen, NULL);
		free(qe);
	}
	else if (f) {
		a = (AV *)SvRV(f);
		if (!(fq = raah2fss_query(a))) {
			free_vecs(q, len);
			free(q);
			return NULL;
		}
		dp = fss_wsh(q, len, w, mode, weight_type, &nd, &total, NULL, fq);
		// free(fq); XXX
	}
	else {
		dp = wsh(q, len, w, mode, weight_type, &nd, &total, NULL);
	}
	free_vecs(q, len);
	free(q);

	if (!dp) {
		return NULL;
	}

	qsort(dp, nd, sizeof *dp, sym_compar);
	if (tak_iw_limit) {
		if (wt = xs_wam_desc(target_wam)) {
			nd = iwl(dp, nd, tak_iw_limit, wt, WAM_REVERT_DIRECTION(mode));
		}
	}
/*
 * lastnd no atai ha, perl karaha &wam::get_last_nd de toreru.
 * wsa deha, ws to ha kotornari, lastnd ha nd to onazi ni naru.
 */
	lastnd = total;
	r = s2rah(dp, nd, w, WAM_REVERT_DIRECTION(mode));

	free(dp);
	return r;
}

static SV *
ncsb0(s, cs_type, d, mode, weight_type, elemn, cno, cswmax)
SV *s;
int cs_type;
int d;
int mode;
int weight_type;
size_t elemn;
int32_t cno;
size_t cswmax;
{
	AV *a;
	syminfo *q = NULL;
	ssize_t len;
	WAM *w;
	AV *na;
	int32_t i;
	struct cs_elem *cc;

	if (!(SvROK(s) && SvTYPE(SvRV(s)) == SVt_PVAV)) {
		return NULL;
	}
	a = (AV *)SvRV(s);
	if (!(w = xs_wam_desc(d))) {
		return NULL;
	}
	if (!(q = rah2s(a, w, mode, &len))) {
		return NULL;
	}

	cc = ncsb(q, len, cs_type, w, mode, weight_type, elemn, &cno, weight_type, cswmax);
	free_vecs(q, len);
	free(q);

	if (!cc) {
		return NULL;
	}

	na = newAV();
	av_extend(na, cno);
	for (i = 0; i < cno; i++) {
		SV *r;
		syminfo *dp;
		size_t nd;
		dp = cc[i].csa.s;
		nd = cc[i].csa.n;
		qsort(dp, nd, sizeof *dp, sym_compar);
		r = s2rah(dp, nd, w, mode);
		av_push(na, r);
		free(dp);
		free(cc[i].csw.s);
	}

	free(cc);
	return newRV_noinc((SV *)na);
}

static SV *
idl2namel0(d, mode, ra)
int d;
int mode;
SV *ra;
{
	size_t len;
	size_t i;
	AV *na;
	AV *a;
	WAM *w;
	SV *r;

	r = NULL;
	if (SvROK(ra) && SvTYPE(SvRV(ra)) == SVt_PVAV && (w = xs_wam_desc(d))) {
		a = (AV *)SvRV(ra);
		na = newAV();
		len = av_len(a) + 1;
		av_extend(na, len);
		for (i = 0; i < len; i++) {
			SV **sp;
			sp = av_fetch(a, i, 0);
			if (sp && SvIOK(*sp)) {
				char const *p;
				SV *sv = NULL;
				idx_t id;
				id = SvIV(*sp);
				if ((p = wam_id2name(w, mode, id)) != NULL) {
					sv = newSVpv(p, 0);
				}
				av_push(na, sv);
			}
		}
		r = newRV_noinc((SV *)na);
	}
	return r;
}

static SV *
namel2idl0(d, mode, ra)
int d;
int mode;
SV *ra;
{
	size_t len;
	size_t i;
	AV *na;
	AV *a;
	WAM *w;
	SV *r;

	r = NULL;
	if (SvROK(ra) && SvTYPE(SvRV(ra)) == SVt_PVAV && (w = xs_wam_desc(d))) {
		a = (AV *)SvRV(ra);
		na = newAV();
		len = av_len(a) + 1;
		av_extend(na, len);
		for (i = 0; i < len; i++) {
			SV **sp;
			sp = av_fetch(a, i, 0);
			if (sp && SvPOK(*sp)) {
				SV *sv;
				char *name;
				STRLEN l;
				idx_t id;
				name = SvPV(*sp, l);
				id = wam_name2id(w, mode, name);
				sv = newSViv(id);
				av_push(na, sv);
			}
		}
		r = newRV_noinc((SV *)na);
	}
	return r;
}

static void
free_vecs(q, len)
syminfo *q;
size_t len;
{
	size_t i;
	for (i = 0; i < len; i++) {
		free(q[i].v);
	}
}

MODULE = wam		PACKAGE = wam

int
constant(name, arg)
	char *name
	int arg


int
init(projectroot)
	char *projectroot

	CODE:
	if (strEQ(projectroot, "")) {
		projectroot = NULL;
	}
	RETVAL = wam_init(projectroot);

	OUTPUT:
	RETVAL


int
open(handle)
	char *handle

	CODE:
	RETVAL = xs_wam_open(handle);

	OUTPUT:
	RETVAL


int
close(d)
	int d;

	CODE:
	RETVAL = xs_wam_close(d);

	OUTPUT:
	RETVAL


ssize_t
size(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_size(w, mode);
	}

	OUTPUT:
	RETVAL

ssize_t
elem_num(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	WAM *w;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_elem_num(w, mode, idx);
	}

	OUTPUT:
	RETVAL


ssize_t
freq_sum(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	WAM *w;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_freq_sum(w, mode, idx);
	}

	OUTPUT:
	RETVAL


ssize_t
max_elem_num(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_max_elem_num(w, mode);
	}

	OUTPUT:
	RETVAL

ssize_t
max_freq_sum(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_max_freq_sum(w, mode);
	}

	OUTPUT:
	RETVAL


ssize_t
total_elem_num(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_total_elem_num(w, mode);
	}

	OUTPUT:
	RETVAL


ssize_t
total_freq_sum(d, mode)
	int d;
	int mode;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = -1;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_total_freq_sum(w, mode);
	}

	OUTPUT:
	RETVAL


SV *
id2name(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	char const *name;
	WAM *w;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = NULL;
	if ((w = xs_wam_desc(d)) && (name = wam_id2name(w, mode, idx)) != NULL) {
		RETVAL = newSVpv(name, 0);
	}

	OUTPUT:
	RETVAL


size_t
name2id(d, mode, name)
	int d;
	int mode;
	char *name;

	PREINIT:
	WAM *w;

	CODE:
	RETVAL = 0;
	if (w = xs_wam_desc(d)) {
		RETVAL = wam_name2id(w, mode, name);
	}

	OUTPUT:
	RETVAL


SV *
get_vec(d, mode, id)
	int d;
	int mode;
	size_t id;

	PREINIT:
	ssize_t e;
	struct xr_vec const *vec;
	WAM *w;
	idx_t idx;

        CODE:
	idx = id;
	RETVAL = NULL;
	if ((w = xs_wam_desc(d)) && (e = wam_get_vec(w, mode, idx, &vec)) != -1) {
		RETVAL = vec2rh(vec, w, WAM_REVERT_DIRECTION(mode));
	}

	OUTPUT:
	RETVAL


SV *
wsh(s, d, mode, weight_type, maxnd, tak_iw_limit, target_wam)
	SV *s;
	int d;
	int mode;
	int weight_type;
	size_t maxnd;
	size_t tak_iw_limit;
	int target_wam;

	CODE:
	RETVAL = wsh0(s, NULL, NULL, d, mode, weight_type, maxnd, tak_iw_limit, target_wam);

	OUTPUT:
	RETVAL


SV *
wshb(s, e, d, mode, weight_type, maxnd, tak_iw_limit, target_wam)
	SV *s;
	SV *e;
	int d;
	int mode;
	int weight_type;
	size_t maxnd;
	size_t tak_iw_limit;
	int target_wam;

	CODE:
	RETVAL = wsh0(s, e, NULL, d, mode, weight_type, maxnd, tak_iw_limit, target_wam);

	OUTPUT:
	RETVAL


SV *
wshf(s, f, d, mode, weight_type, maxnd, tak_iw_limit, target_wam)
	SV *s;
	SV *f;
	int d;
	int mode;
	int weight_type;
	size_t maxnd;
	size_t tak_iw_limit;
	int target_wam;

	CODE:
	RETVAL = wsh0(s, NULL, f, d, mode, weight_type, maxnd, tak_iw_limit, target_wam);

	OUTPUT:
	RETVAL


SV *
ncsb(s, cs_type, d, mode, weight_type, elemn, cno, cswmax)
	SV *s;
	int cs_type;
	int d;
	int mode;
	int weight_type;
	size_t elemn;
	size_t cno;
	size_t cswmax;

	CODE:
	RETVAL = ncsb0(s, cs_type, d, mode, weight_type, elemn, cno, cswmax);

	OUTPUT:
	RETVAL


size_t
get_last_nd()

	CODE:
	RETVAL = lastnd;

	OUTPUT:
	RETVAL


SV *
getprop(d, dir, key, id)
	int d;
	int dir;
	char *key;
	size_t id;

	PREINIT:
	char const *p;
	WAM *w;
	WAM *f;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = NULL;
	if ((w = xs_wam_desc(d)) && (f = wam_prop_open(w, dir, key)) != NULL) {
		if (wam_prop_gets(f, idx, &p)>0) {
			RETVAL = newSVpv(p, 0);
		}
		wam_close(f);
	}

	OUTPUT:
	RETVAL


SV *
getfss(d, id, segid)
	int d;
	size_t id;
	int segid;

	PREINIT:
	char const *p;
	WAM *w;
	WAM *f;
	idx_t idx;

	CODE:
	idx = id;
	RETVAL = NULL;
	if ((w = xs_wam_desc(d)) && (f = wam_fss_open(w)) != NULL) {
		if (wam_fss_gets(f, idx, segid, &p)>0) {
			RETVAL = newSVpv(p, 0);
		}
		wam_close(f);
	}

	OUTPUT:
	RETVAL


SV *
text2vec(text, d, mode, stemmer)
	char *text;
	int d;
	int mode;
	char *stemmer;

	PREINIT:
	struct xr_vec *vec;
	WAM *w;

	CODE:
	RETVAL = NULL;

	if ((w = xs_wam_desc(d)) && (vec = text2vec(text, w, mode, stemmer))) {
		RETVAL = vec2rh(vec, w, mode);
		free(vec);
	}

	OUTPUT:
	RETVAL


SV *
get_weight_types()
	PREINIT:
	size_t i;
	AV *na;

	CODE:
	na = newAV();
	for (i = 0; i < nelems(weight_types); i++) {
		SV *sv;
		sv = newSVpv(weight_types[i], 0);
		av_push(na, sv);
	}
	RETVAL = newRV_noinc((SV *)na);

	OUTPUT:
	RETVAL


SV *
idl2namel(d, mode, ra)
	int d;
	int mode;
	SV *ra;

	CODE:
	RETVAL = idl2namel0(d, mode, ra);

	OUTPUT:
	RETVAL


SV *
namel2idl(d, mode, ra)
	int d;
	int mode;
	SV *ra;

	CODE:
	RETVAL = namel2idl0(d, mode, ra);

	OUTPUT:
	RETVAL

